"""
======
Server
======

This module manages loading the modbus server or Slave
"""
import logging

from pymodbus.compat import get_next, iteritems, iterkeys
from pymodbus.datastore import ModbusServerContext, ModbusSlaveContext
from pymodbus.datastore.store import BaseModbusDataBlock
from pymodbus.device import ModbusDeviceIdentification
from pymodbus.exceptions import ParameterException
from pymodbus.server.asynchronous import (
    StartSerialServer,
    StartTcpServer,
    StartUdpServer,
    StopServer,
)
from pymodbus.version import version
from twisted.internet.task import LoopingCall

DEFAULTCONF = {
    "tcp": {"context": None, "identity": None, "address": ("", 502), "ignore_missing_slaves": False},
    "udp": {"context": None, "identity": None, "address": ("", 502), "ignore_missing_slaves": False},
    "rtu": {
        "context": None,
        "identity": None,
        "port": "/dev/ttyS0",
        "buadrate": 19200,
        "bytesize": 8,
        "stopbits": 1,
        "parity": "N",
        "timeout": 0,
        "xonxoff": 0,
        "rtscts": 0,
    },
}

log = logging.getLogger(__name__)
log.debug("pymodbus version: %s", version.short())


class ModbusSparseDataBlock(BaseModbusDataBlock):
    """
    Cloned of ModbusSparseDataBlock from pymodbus 2.5.3 to backport for python vesions 3.5 and 3.6

    Creates a sparse modbus datastore
    E.g Usage.
    sparse = ModbusSparseDataBlock({10: [3, 5, 6, 8], 30: 1, 40: [0]*20})
    This would create a datablock with 3 blocks starting at
    offset 10 with length 4 , 30 with length 1 and 40 with length 20
    sparse = ModbusSparseDataBlock([10]*100)
    Creates a sparse datablock of length 100 starting at offset 0 and default value of 10
    sparse = ModbusSparseDataBlock() --> Create Empty datablock
    sparse.setValues(0, [10]*10)  --> Add block 1 at offset 0 with length 10 (default value 10)
    sparse.setValues(30, [20]*5)  --> Add block 2 at offset 30 with length 5 (default value 20)
    if mutable is set to True during initialization, the datablock can not be altered with
    setValues (new datablocks can not be added)
    """

    def __init__(self, values=None, mutable=True):
        """
        Initializes a sparse datastore. Will only answer to addresses
        registered, either initially here, or later via setValues()
        :param values: Either a list or a dictionary of values
        :param mutable: The data-block can be altered later with setValues(i.e add more blocks)
        If values are list , This is as good as sequential datablock.
        Values as dictionary should be in {offset: <values>} format, if values
        is a list, a sparse datablock is created starting at offset with the length of values.
        If values is a integer, then the value is set for the corresponding offset.
        """
        self.values = {}
        self._process_values(values)
        self.mutable = mutable
        self.default_value = self.values.copy()
        self.address = get_next(iterkeys(self.values), None)

    @classmethod
    def create(klass, values=None):
        """Factory method to create sparse datastore.
        Use setValues to initialize registers.
        :param values: Either a list or a dictionary of values
        :returns: An initialized datastore
        """
        return klass(values)

    def reset(self):
        """Reset the store to the initially provided defaults"""
        self.values = self.default_value.copy()

    def validate(self, address, count=1):
        """Checks to see if the request is in range
        :param address: The starting address
        :param count: The number of values to test for
        :returns: True if the request in within range, False otherwise
        """
        if count == 0:
            return False
        handle = set(range(address, address + count))
        return handle.issubset(set(iterkeys(self.values)))

    def getValues(self, address, count=1):
        """Returns the requested values of the datastore
        :param address: The starting address
        :param count: The number of values to retrieve
        :returns: The requested values from a:a+c
        """
        return [self.values[i] for i in range(address, address + count)]

    def _process_values(self, values):
        def _process_as_dict(values):
            for idx, val in iteritems(values):
                if isinstance(val, (list, tuple)):
                    for i, v in enumerate(val):
                        self.values[idx + i] = v
                else:
                    self.values[idx] = int(val)

        if isinstance(values, dict):
            _process_as_dict(values)
            return
        if hasattr(values, "__iter__"):
            values = dict(enumerate(values))
        elif values is None:
            values = {}  # Must make a new dict here per instance
        else:
            raise ParameterException("Values for datastore must " "be a list or dictionary")
        _process_as_dict(values)

    def setValues(self, address, values, use_as_default=False):
        """Sets the requested values of the datastore
        :param address: The starting address
        :param values: The new values to be set
        :param use_as_default: Use the values as default
        """
        if isinstance(values, dict):
            new_offsets = list(set(list(values.keys())) - set(list(self.values.keys())))
            if new_offsets and not self.mutable:
                raise ParameterException("Offsets {} not " "in range".format(new_offsets))
            self._process_values(values)
        else:
            if not isinstance(values, list):
                values = [values]
            for idx, val in enumerate(values):
                if address + idx not in self.values and not self.mutable:
                    raise ParameterException("Offset {} not " "in range".format(address + idx))
                self.values[address + idx] = val
        if not self.address:
            self.address = get_next(iterkeys(self.values), None)
        if use_as_default:
            for idx, val in iteritems(self.values):
                self.default_value[idx] = val


class Task(object):
    """
    A Class to manage all the task to run periodically alongside the server

    This Can be called as a decorators.  The following example will print "Hello\\nworld" every second.
    The decorator can be used multiple times to call more than once.

    ..
        @mbsim.core.server.Task(1, args=("Hello", "world"))
        def hw(*args):
            for arg in args:
                print(arg)

        or

        mbsim.core.server.Task(1, func=hw, args=("Hello", "world"))

    Both example are equivalent
    """

    tasks = []

    def __init__(self, inter, func=None, args=(), kwargs={}, now=False):
        """
        This will take the arguments for the task to be run

        :param inter: Interval of time to wait before calling the no blocking function
        :type inter: float
        :param func: Function to call
        :param args: arguments to be past to task
        :type args: tuple, optional
        :param kwargs: key word arguments to be passed to task, defaults to {}
        :type kwargs: dict, optional
        :param now: To run immediately or wait the interval, defaults to False
        :type now: bool, optional
        """
        self.inter = inter
        self.args = args
        self.kwargs = kwargs
        self.now = now
        self.tasks.append(self)
        if func:
            self(func)

    def __call__(self, func):
        """
        Init the task and return original function

        :param func: A Non blocking function
        :type func: function
        :return: Original function
        :rtype: function
        """
        self.func = func
        log.debug("Adding task: %s", (self.func.__name__, self.inter, self.args, self.kwargs))
        self.task = LoopingCall(f=self.wrap, a=self.args, kw=self.kwargs)
        return func

    def wrap(self, a, kw):
        """
        A wrap for twisted to pass a and kw as args.  This function will explode them into the function

        :param a: args
        :type a: tuple
        :param kw: kwargs
        :type kw: dict
        """
        return self.func(*a, **kw)

    def start(self):
        """
        To start the task
        """
        log.debug("Starting task: %s", (self.func.__name__, self.inter, self.args, self.kwargs))
        self.task.start(self.inter, now=self.now)

    @classmethod
    def startTasks(cls):
        """
        Class method to start all tasks
        """
        for task in cls.tasks:
            task.start()

    def stop(self):
        """
        To stop a task
        """
        log.debug("Stopping task: %s", (self.func.__name__, self.inter, self.args, self.kwargs))
        self.task.stop()

    @classmethod
    def stopTasks(cls):
        """
        Class method to stop all tasks
        """
        for task in cls.tasks:
            task.stop()


def genDevice(**kwargs):
    """
    A function to create a identity.  All parameters are key word arguments.

    :param name: Vendor Name. Defaults to `mbsim`
    :type name: str
    :param code: Product Code. Defaults to `MB`
    :type code: str
    :param url: Vendor URL. Defaults to `https://gitlab.com/nee2c/mbsim`
    :type url: str
    :param product: Product Name. Defaults to `mbsim`
    :type product: str
    :param model: Model Name. Defaults to `mbsim`
    :type model: str
    :param version: Version. Defaults to `1.0.0`
    :type version: str
    :return: Returns an identity for server
    """
    identity = ModbusDeviceIdentification()

    identity.VendorName = kwargs.get("name", "mbsim")
    identity.ProductCode = kwargs.get("code", "MB")
    identity.VendorUrl = kwargs.get("url", "https://gitlab.com/nee2c/mbsim")
    identity.ProductName = kwargs.get("product", "mbsim")
    identity.ModelName = kwargs.get("model", "mbsim")
    identity.MajorMinorRevision = kwargs.get("version", "1.0.0")

    log.debug("Genrated identity: %s", identity.summary())
    return identity


def genContext(context=None, single=True):
    """
    A Function to return Context for Server.

    The context can be None. This will genrate a context that all slavesid will use. All values set to 0

    If context is an instance of ModbusServerContext, funtion returns the context

    Else dict of slaves with dict of address space("di", "co", "hr", "ir") with list or dict {offset: [reg0, reg1, ...]}


    dict Context example

    ..

        {0: {"di": [0, 1, 0, 1, 0], "co": {0: [1, 0, 1], 100: [1]}, "hr": [123, 10], "ir": {123: [123, 0, 123]}}, ...}

    :param context: Server Context or dict of Slaves context or dict
    :param single:
    :type single: bool
    """
    if isinstance(context, ModbusServerContext):
        log.debug("Already a Server Context")
        return context
    if context is None:
        log.debug("Genirating Server context: single: %s", single)
        return ModbusServerContext(slaves=ModbusSlaveContext(zero_mode=True), single=single)
    slaves = {}
    log.debug("Genrating context from %s", context)
    for slaveid, slavecontext in context.items():
        slaves[slaveid] = ModbusSlaveContext(
            **{key: ModbusSparseDataBlock(values=vals) for key, vals in slavecontext.items()},
            zero_mode=True,
        )
    return ModbusServerContext(slaves=slaves, single=single)


def start(server, **kwargs):
    """
    This is the function to start modbus server.  Passes all kwargs though to server, if missing uses default.

    :param server: The protocal to run modbus server
    :type server: str
    """
    if server not in DEFAULTCONF.keys():
        raise NotImplementedError("This server type {} is not Iplemented".format(server))
    serverargs = {**kwargs, **{key: val for key, val in DEFAULTCONF[server].items() if key not in kwargs.keys()}}
    if not serverargs.get("context"):
        serverargs["context"] = genContext()
    log.info("Starting %s server", server)
    log.debug("Server args: %s", str(serverargs))

    log.debug("Starting Tasks")
    Task.startTasks()
    log.debug("Started Tasks")

    log.debug("Starting %s Server", server)
    if server == "tcp":
        StartTcpServer(**serverargs)
    elif server == "udp":
        StartUdpServer(**serverargs)
    elif server == "rtu":
        StartSerialServer(**serverargs)
    log.debug("Started Server")


def stop():
    """
    Function to stop modbus server
    """
    log.debug("Stopping Tasks")
    Task.stopTasks()
    log.debug("Stopped Tasks")
    log.debug("Stopping Server")
    StopServer()
    log.debug("Stopped Server")



"""
Test for mbsim-core.server module
"""
import builtins
from importlib import import_module, reload
from itertools import combinations

import pytest
from pymodbus.datastore import ModbusServerContext, ModbusSlaveContext

import mbsim.core.server as mbserver


@pytest.fixture(scope="class")
def mb():
    """
    Import lib for testing and reload
    """
    mb = import_module("mbsim.core.server")
    yield mb
    reload(mb)


def genDevData():
    """
    Genrates device identity test data
    """
    default = {
        "VendorName": "mbsim",
        "ProductCode": "MB",
        "VendorUrl": "https://gitlab.com/nee2c/mbsim",
        "ProductName": "mbsim",
        "ModelName": "mbsim",
        "MajorMinorRevision": "1.0.0",
    }
    alt = {
        "VendorName": "tests",
        "ProductCode": "AZ",
        "VendorUrl": "https://example.com",
        "ProductName": "sim",
        "ModelName": "mb",
        "MajorMinorRevision": "1.0.1",
    }
    paramname = {
        "VendorName": "name",
        "ProductCode": "code",
        "VendorUrl": "url",
        "ProductName": "product",
        "ModelName": "model",
        "MajorMinorRevision": "version",
    }
    data = [(None, default)]
    for num, _ in enumerate(default, 1):
        for keys in combinations(default.keys(), num):
            test = {paramname[key]: alt[key] for key in keys}
            res = {
                **{key: alt[key] for key in keys},
                **{key: default[key] for key in set(default.keys()).difference(keys)},
            }
            data.append((test, res))
    return data


def genContext():
    return mbserver.genContext()


class TestGenDevice(object):
    """
    This is group test for the function of genDevice
    """

    @pytest.mark.parametrize("test,res", genDevData())
    def test_genDevice(self, test, res, mb):
        """
        Test identity after it genrated
        """
        if test is None:
            test = {}
        dev = mb.genDevice(**test)
        for key, value in res.items():
            assert getattr(dev, key) == value


class TestGenContext(object):
    """
    This is a group of text for testing generating Context
    """

    _fx = {2: "d", 4: "i", **{i: "h" for i in (3, 6, 16, 22, 23)}, **{i: "c" for i in (1, 5, 15)}}

    def test_passingNone(self, mb):
        """
        test calling genContext with no arguments
        """
        test = mb.genContext()
        assert isinstance(test, ModbusServerContext)
        assert test.single
        assert test[0] is test[1]
        for fx in self._fx.keys():
            assert test[0].getValues(fx, 200, count=5)

    def test_passingContext(self, mb):
        """
        test passing the function exsisting Server Context
        """
        test = ModbusServerContext()
        assert mb.genContext(test) is test

    @pytest.mark.parametrize(
        "testdict",
        [
            (
                {
                    2: {
                        "co": {0: [1, 0, 0, 1], 101: [0, 1, 1, 0]},
                        "di": {0: [1, 0, 0, 1], 201: [0, 1, 1, 0]},
                        "hr": {60: [100, 580, 40, 1], 101: [0, 101, 1111, 0]},
                        "ir": {10: [1000, 0, 0, 541], 101: [870, 221, 1, 0]},
                    }
                }
            ),
            (
                {
                    2: {
                        "co": [1, 0, 0, 1],
                        "di": [1, 0, 0, 1],
                        "hr": [100, 580, 40, 1],
                        "ir": [1000, 0, 0, 5410, 221, 1, 0],
                    }
                }
            ),
            (
                {
                    2: {
                        "co": [1, 0, 0, 1],
                        "di": [1, 0, 0, 1],
                        "hr": {60: [100, 580, 40, 1], 101: [0, 101, 1111, 0]},
                        "ir": {10: [1000, 0, 0, 541], 101: [870, 221, 1, 0]},
                    }
                }
            ),
            (
                {
                    2: {
                        "co": [1, 0, 0, 1],
                        "di": [1, 0, 0, 1],
                        "hr": [100, 580, 40, 1],
                        "ir": [1000, 0, 0, 5410, 221, 1, 0],
                    },
                    20: {
                        "co": [1, 1, 1, 1],
                        "di": [0, 0, 0, 0],
                        "hr": [107, 587, 44, 6],
                        "ir": [1006, 4, 0, 5402, 221, 4, 0],
                    },
                }
            ),
        ],
    )
    def test_customContext(self, testdict, mb):
        """
        create and test custom setup context
        """
        _fx = {"co": 1, "di": 2, "hr": 3, "ir": 4}
        test = mb.genContext(testdict, single=False)
        assert isinstance(test, ModbusServerContext)
        for slave in test.slaves():
            assert slave in testdict
            assert isinstance(test[slave], ModbusSlaveContext)
            for store, vals in testdict[slave].items():
                if isinstance(vals, dict):
                    for offset, regs in vals.items():
                        assert test[slave].getValues(_fx[store], offset, len(regs)) == regs
                elif isinstance(vals, list):
                    assert test[slave].getValues(_fx[store], 0, len(vals)) == vals


class MockCall(object):
    """
    A simple Test function to count the number of calls and arguments used
    """

    def __init__(self):
        """
        Init the test function
        """
        self.count = 0
        self.args = []
        self.kwargs = []

    def __call__(self, *args, **kwargs):
        """
        Increment the count due to being called
        """
        self.count += 1
        self.args.append(args)
        self.kwargs.append(kwargs)


class MockLoopingCall(object):
    """
    Mock object to mock twisted LoopingCall object
    """

    def __init__(self, f, a=None, kw=None):
        """
        test object to test LoopingCall

        :param f: function
        :param a: args, defaults to None
        :type a: tuple, optional
        :param kw: kwargs, defaults to None
        :type kw: dict, optional
        """
        self.started = False
        self.stopped = False
        self.func = f
        self.args = a
        self.kwargs = kw
        self.inter = None
        self.now = None

    def start(self, inter, now=False):
        """
        mock loopingCall start function

        :param inter: time interval
        :type inter: float
        :param now: first call or wait interval, defaults to False
        :type now: bool, optional
        """
        if self.started:
            raise RuntimeError("Mock Task has already been started")
        self.started = True
        self.inter = inter
        self.now = now

    def stop(self):
        """
        mock loopingCall stop function

        :param inter: time interval
        :type inter: float
        :param now: first call or wait interval, defaults to False
        :type now: bool, optional
        """
        if self.stopped:
            raise RuntimeError("Mock Task has already been started")
        self.stopped = True


class Test_Tasks(object):
    """
    Group of test to test Tasks
    """

    @pytest.mark.parametrize(
        "index,inter,args,kwargs,now,func",
        [
            (0, 1, (), {}, False, print),
            (1, 2, ("Hello"), {}, False, dict),
            (2, 3, (), {"Hello": "World"}, False, set),
            (3, 4, (), {}, True, max),
        ],
    )
    def test_createDecorator(self, index, inter, args, kwargs, now, func, mb, monkeypatch):
        """
        test the creation of tasks
        """
        monkeypatch.setattr(mb, "LoopingCall", MockLoopingCall)
        mb.Task(inter, args=args, kwargs=kwargs, now=now)(func)
        assert func == mb.Task.tasks[index].func
        assert inter == mb.Task.tasks[index].inter
        assert args == mb.Task.tasks[index].args
        assert kwargs == mb.Task.tasks[index].kwargs
        assert now == mb.Task.tasks[index].now
        assert isinstance(mb.Task.tasks[index].task, MockLoopingCall)

    @pytest.mark.parametrize(
        "index,inter,args,kwargs,now,func",
        [
            (0, 1, (), {}, False, print),
            (1, 2, ("Hello"), {}, False, dict),
            (2, 3, (), {"Hello": "World"}, False, set),
            (3, 4, (), {}, True, max),
        ],
    )
    def test_createInstance(self, index, inter, args, kwargs, now, func, mb, monkeypatch):
        """
        test the creation of tasks
        """
        monkeypatch.setattr(mb, "LoopingCall", MockLoopingCall)
        mb.Task(inter, func=func, args=args, kwargs=kwargs, now=now)
        assert func == mb.Task.tasks[index].func
        assert inter == mb.Task.tasks[index].inter
        assert args == mb.Task.tasks[index].args
        assert kwargs == mb.Task.tasks[index].kwargs
        assert now == mb.Task.tasks[index].now
        assert isinstance(mb.Task.tasks[index].task, MockLoopingCall)

    def test_startTasks(self, mb):
        """
        Test class method and by proxy instace start method
        """
        mb.Task.startTasks()
        for task in mb.Task.tasks:
            assert task.task.started
            assert task.inter == task.task.inter
            assert task.now == task.task.now
            assert task.args == task.task.args
            assert task.kwargs == task.task.kwargs

    def test_stopTasks(self, mb):
        """
        Test class method and by proxy instace stop method
        """
        mb.Task.stopTasks()
        for task in mb.Task.tasks:
            assert task.task.stopped

    @pytest.mark.xfail(reason="Monkeypatch fails to patch print in py37 and py38, need to rework MockCall")
    def test_twistedwrap(self, mb, monkeypatch):
        """
        Test wrap function for twisted loopingCall passing args
        """
        task = mb.Task.tasks[0]
        monkeypatch.setattr(builtins, "print", MockCall())
        test = task.wrap(a=("Hello", "World"), kw={"foo": 1, "bar": 2})
        assert test.count == 1
        assert test.args[0] == ("Hello", "World")
        assert test.kwargs[0] == {"foo": 1, "bar": 2}


class Test_Start(object):
    """
    Group of test to test the start function
    """

    def test_notimplemented(self, mb):
        """
        Test if unexpected server raise NotImplementedError
        """
        with pytest.raises(NotImplementedError, match="This server type Hello is not Iplemented"):
            mb.start("Hello")

    @pytest.mark.parametrize("context", [None, genContext()])
    def test_Context(self, context, monkeypatch, mb):
        """
        Test to ckeck behavior whith diffrent types of context is used as prams
        """
        server, mock = "tcp", MockCall()
        monkeypatch.setattr(mb, "StartTcpServer", mock)
        if context is None:
            # Call with no context param
            mb.start(server)
        mb.start(server, context=context)
        if context is None:
            assert mock.count == 2
        else:
            assert mock.count
        for kwargs in mock.kwargs:
            assert isinstance(kwargs.get("context"), ModbusServerContext)

    @pytest.mark.parametrize("test", ["tcp", "udp", "rtu"])
    def test_servercalls(self, test, monkeypatch, mb):
        """
        Test the pymodbus module is called correctly
        """
        servers = {
            "tcp": ("StartTcpServer", MockCall()),
            "udp": ("StartUdpServer", MockCall()),
            "rtu": ("StartSerialServer", MockCall()),
        }
        for server, (attr, mock) in servers.items():
            monkeypatch.setattr(mb, attr, mock)
        mb.start(test)
        _, res = servers.pop(test)
        assert res.count == 1
        for _, mock in servers.values():
            assert not mock.count

    def test_startTasks(self, mb, monkeypatch):
        """
        Test to make sure Tasks are started
        """
        test = MockCall()
        monkeypatch.setattr(mb, "StartTcpServer", MockCall())
        monkeypatch.setattr(mb.Task, "startTasks", test)
        mb.start("tcp")
        assert test.count == 1


class TestStop(object):
    """
    Test Stop functions
    """

    def test_stop(self, monkeypatch, mb):
        """
        Test Stop functions
        """
        mock = MockCall()
        monkeypatch.setattr(mb, "StopServer", mock)
        mb.stop()
        assert mock.count == 1
        assert not mock.args[0]
        assert not mock.kwargs[0]

    def test_stopTasks(self, mb, monkeypatch):
        """
        Test to make sure Tasks are started
        """
        test = MockCall()
        monkeypatch.setattr(mb, "StopServer", MockCall())
        monkeypatch.setattr(mb.Task, "stopTasks", test)
        mb.stop()
        assert test.count == 1
